# Name:       SubStanding for Assetto Corsa
# Version:    v1.1
# Anthor:     Sylvain Villet
# Contact:    sylvain.villet@gmail.com
# Date:       22.09.2017
# Desc.:      This app provides standings and positions
#             in a subgroup of the cars, that can be
#             the same car as the player or a class like
#             LMP1, GT2, GT3 etc defined here.
#             It also provides the global standing with
#             the same design (max 36 cars).
#             The app can be fully configured from the SubStanding_config widget.
#
#             Feel free to add, remove or modify the "classes"
#             definitions in the Substanding_classes/classes.ini file.
#
# Thanks:     - Rombik for the sim_info module
#             - Rivali (OV1Info) and Fernando Deutsch (ferito-LiveCarTracker)
#             for the inspiration and example
#             - ExOAte for the beta tests and ideas

import ac
import acsys
import sys
import os
import configparser
import time
import platform
import math
import traceback

# Global
updateTime = 200

# Standings
showLogo = 0
showTitle = 0
fontSize = 16
opacity = 50
bottomAnchor = -1
showBorder = 1
showBadge = 1
showTyres = 1
showPitStops = 1
showDelta = 1
showLapDiffRelative = 1
colorAt = 250
unit = "metric"
driverDisplayedGlobal = 3
driverDisplayedRelative = 3
driverDisplayedClass = 3
driverDisplayedSameCar = 3

# Positions
showLogoPos = 0
showTitlePos = 1
fontSizePos = 32
opacityPos = 50
showBorderPos = 1

#Classes from the classes.ini file
classes = []

#Teammates and rivals from the drivers.ini file
teammates = []
rivals = []

serverIp = ""

# Global variables
spacing = 5
firstSpacing = 30
fontSizeConfig = 16

maxDriverCount = 32

driversFullList =[]
myDriver = 0
myCarName = 0
myClass = ["No class"]
trackLength = 0
lastUpdateTime = 0

badge=[]

debug = 0
cpuUsage = 0

# Objects
SubStandings = {
    "model": 0,
    "class": 0,
    "global": 0,
    "relative": 0
}

SubPositions = {
    "model": 0,
    "class": 0,
    "global": 0
}

BottomAnchors = {
    "SubStanding_Global": 0,
    "SubStanding_Relative": 0,
    "SubStanding_SameCar": 0,
    "SubStanding_Class": 0
}

config = 0
classesConfig = 0
configApp = 0

# import libraries
try:
    if platform.architecture()[0] == "64bit":
      sysdir='apps/python/SubStanding/SubStanding_dll/stdlib64'
    else:
      sysdir='apps/python/SubStanding/SubStanding_dll/stdlib'
    sys.path.insert(0, sysdir)
    os.environ['PATH'] = os.environ['PATH'] + ";."

    from SubStanding_lib.sim_info import info
except:
    ac.log("SubStanding: Error importing libraries: %s" % traceback.format_exc())
    raise

def acMain(ac_version):
    try:
        global myCarName, myClass, trackLength, cpuUsage
        global config, classesConfig, configApp
        global updateTime
        global showLogo, showTitle, fontSize, opacity, bottomAnchor, showBorder, showDelta, showLapDiffRelative, colorAt, unit
        global driverDisplayedGlobal, driverDisplayedRelative, driverDisplayedClass, driverDisplayedSameCar
        global showLogoPos, showTitlePos, fontSizePos, opacityPos, showBorderPos
        global badge, showBadge, showTyres, showPitStops
        global teammates, rivals, maxDriverCount
        global myDriver, driversFullList

        config = configparser.ConfigParser()
        config.read("apps/python/SubStanding/SubStanding_config/config.ini")

        # Global
        updateTime = config.getint("GLOBAL", "updateTime")

        # Standings
        showLogo  = config.getint("STANDINGS", "showLogo")
        showTitle = config.getint("STANDINGS", "showTitle")
        fontSize  = config.getint("STANDINGS", "fontSize")
        opacity   = config.getint("STANDINGS", "opacity")
        bottomAnchor = config.getint("STANDINGS", "bottomAnchor")
        showBorder = config.getint("STANDINGS", "showBorder")
        showBadge = config.getint("STANDINGS", "showBadge")
        showTyres = config.getint("STANDINGS", "showtyres")
        showPitStops = config.getint("STANDINGS", "showpitstops")
        showDelta = config.getint("STANDINGS", "showDelta")
        showLapDiffRelative = config.getint("STANDINGS", "showLapDiffRelative")
        colorAt   = config.getint("STANDINGS", "colorAt")
        unit      = config.get("STANDINGS", "unit")
        driverDisplayedGlobal  = config.getint("STANDINGS", "driverDisplayedGlobal")
        driverDisplayedRelative  = config.getint("STANDINGS", "driverDisplayedRelative")
        driverDisplayedClass   = config.getint("STANDINGS", "driverDisplayedClass")
        driverDisplayedSameCar = config.getint("STANDINGS", "driverDisplayedSameCar")

        # Positions
        showLogoPos  = config.getint("POSITIONS", "showLogoPos")
        showTitlePos = config.getint("POSITIONS", "showTitlePos")
        fontSizePos  = config.getint("POSITIONS", "fontSizePos")
        opacityPos   = config.getint("POSITIONS", "opacityPos")
        showBorderPos = config.getint("POSITIONS", "showBorderPos")
        BottomAnchors["SubStanding_Global"] = config.getint("POSITIONS", "bottomAnchorGlobal")
        BottomAnchors["SubStanding_Relative"] = config.getint("POSITIONS", "bottomAnchorRelative")
        BottomAnchors["SubStanding_SameCar"] = config.getint("POSITIONS", "bottomAnchorSameCar")
        BottomAnchors["SubStanding_Class"] = config.getint("POSITIONS", "bottomAnchorClass")

        # Get classes
        classesConfig = configparser.ConfigParser()
        classesConfig.read("apps/python/SubStanding/SubStanding_classes/classes.ini")

        for eachSection in classesConfig.sections():
            classTmp = []
            classTmp.append(eachSection)
            for eachItem in classesConfig.items(eachSection):
                classTmp.append(eachItem[1])
            classes.append(classTmp)

        myCarName = ac.getCarName(0)

        for eachClass in classes:
            if myCarName in eachClass:
                myClass = eachClass
                break

        maxDriverCount = ac.getCarsCount()

        # Get badges
        for i in range(maxDriverCount):
            carName = ac.getCarName(i)
            if type(carName) == type(""):
                textureId = ac.newTexture("content/cars/" + carName + "/ui/badge.png")
                if textureId >= 0:
                    badge.append(textureId)
                else:
                    badge.append(-1)
            else:
                badge.append(-1)

        # Get teammates and rivals
        driversConfig = configparser.ConfigParser()
        driversConfig.read("apps/python/SubStanding/SubStanding_drivers/drivers.ini")

        for eachSection in driversConfig.sections():
            if eachSection == "TEAMMATES":
                for eachItem in driversConfig.items(eachSection):
                    teammates.append(eachItem[1])
            elif eachSection == "RIVALS":
                for eachItem in driversConfig.items(eachSection):
                    rivals.append(eachItem[1])

        SubStandings["global"] = StandingWindow("SubStanding_Global", "Global", driverDisplayedGlobal, 0)
        SubStandings["relative"] = StandingWindow("SubStanding_Relative", "Relative", driverDisplayedRelative, 1)
        SubStandings["model"]  = StandingWindow("SubStanding_SameCar", "Same Car", driverDisplayedSameCar, 0)
        SubStandings["class"]  = StandingWindow("SubStanding_Class", myClass[0], driverDisplayedClass, 0)

        SubPositions["model"]  = PositionWindow("SubStanding_SameCar_Pos", "Same Car")
        SubPositions["class"]  = PositionWindow("SubStanding_Class_Pos", myClass[0])
        SubPositions["global"] = PositionWindow("SubStanding_Global_Pos", "Global")

        configApp = SubStanding_config("SubStanding_config", "SubStanding config", fontSizeConfig)
        configApp.updateView()

        ac.addRenderCallback(SubStandings["model"].window, onRenderCallbackStandingModel)
        ac.addRenderCallback(SubStandings["class"].window, onRenderCallbackStandingClass)
        ac.addRenderCallback(SubStandings["global"].window, onRenderCallbackStandingGlobal)
        ac.addRenderCallback(SubStandings["relative"].window, onRenderCallbackStandingRelative)

        ac.addRenderCallback(SubPositions["model"].window, onRenderCallbackPositionModel)
        ac.addRenderCallback(SubPositions["class"].window, onRenderCallbackPositionClass)
        ac.addRenderCallback(SubPositions["global"].window, onRenderCallbackPositionGlobal)

        ac.addRenderCallback(configApp.window, onRenderCallbackConfig)

        trackLength = ac.getTrackLength(0)

        # Get all needed infos
        for index in range(maxDriverCount):
            driver = {
                "carId":        index,
                "connected":    ac.isConnected(index),
                "driverName":   ac.getDriverName(index),
                "carName":      ac.getCarName(index),
                "bestLap":      ac.getCarState(index, acsys.CS.BestLap),
                "lapCount":     ac.getCarState(index, acsys.CS.LapCount),
                "speedMS":      ac.getCarState(index, acsys.CS.SpeedMS),
                "distance":     ac.getCarState(index, acsys.CS.LapCount) + ac.getCarState(index, acsys.CS.NormalizedSplinePosition),
                "lapDistance":  ac.getCarState(index, acsys.CS.NormalizedSplinePosition),
                "tyres":        ac.getCarTyreCompound(index),
                "isInPitLine":  ac.isCarInPitline(index),
                "pitStops":     0,
                "pitEntryFlag": False,
                "raceStartFlag": True
            }
            if index == 0:
                myDriver = driver

            driversFullList.append(driver)

        if debug:
            cpuUsage = CpuUsage("SubStanding_CpuUsage", "CPU Usage", fontSize)

        return "SubStanding"
    except:
        ac.log("SubStanding: Error in acMain: %s" % traceback.format_exc())


def writeParameters():
    try:
        config.set("GLOBAL", "updateTime", str(updateTime))

        config.set("STANDINGS", "showLogo", str(showLogo))
        config.set("STANDINGS", "showTitle", str(showTitle))
        config.set("STANDINGS", "fontSize", str(fontSize))
        config.set("STANDINGS", "opacity", str(opacity))
        config.set("STANDINGS", "bottomAnchor", str(bottomAnchor))
        config.set("STANDINGS", "showBorder", str(showBorder))
        config.set("STANDINGS", "showBadge", str(showBadge))
        config.set("STANDINGS", "showTyres", str(showTyres))
        config.set("STANDINGS", "showPitStops", str(showPitStops))
        config.set("STANDINGS", "showDelta", str(showDelta))
        config.set("STANDINGS", "showLapDiffRelative", str(showLapDiffRelative))
        config.set("STANDINGS", "colorAt", str(colorAt))
        config.set("STANDINGS", "unit", str(unit))

        config.set("STANDINGS", "driverDisplayedGlobal", str(driverDisplayedGlobal))
        config.set("STANDINGS", "driverDisplayedRelative", str(driverDisplayedRelative))
        config.set("STANDINGS", "driverDisplayedClass", str(driverDisplayedClass))
        config.set("STANDINGS", "driverDisplayedSameCar", str(driverDisplayedSameCar))

        config.set("POSITIONS", "showLogoPos", str(showLogoPos))
        config.set("POSITIONS", "showTitlePos", str(showTitlePos))
        config.set("POSITIONS", "fontSizePos",  str(fontSizePos))
        config.set("POSITIONS", "opacityPos",  str(opacityPos))
        config.set("POSITIONS", "showBorderPos", str(showBorderPos))

        config.set("POSITIONS", "bottomAnchorGlobal", str(BottomAnchors["SubStanding_Global"]))
        config.set("POSITIONS", "bottomAnchorRelative", str(BottomAnchors["SubStanding_Relative"]))
        config.set("POSITIONS", "bottomAnchorSameCar", str(BottomAnchors["SubStanding_SameCar"]))
        config.set("POSITIONS", "bottomAnchorClass", str(BottomAnchors["SubStanding_Class"]))

        configFile = open("apps/python/SubStanding/SubStanding_config/config.ini", 'w')
        config.write(configFile)
        configFile.close()

    except:
        ac.log("SubStanding: Error in writeParameters: %s" % traceback.format_exc())

def refreshAndWriteParameters():
    try:
        SubStandings["model"].refreshParameters(driverDisplayedSameCar)
        SubStandings["class"].refreshParameters(driverDisplayedClass)
        SubStandings["global"].refreshParameters(driverDisplayedGlobal)
        SubStandings["relative"].refreshParameters(driverDisplayedRelative)

        SubPositions["model"].refreshParameters()
        SubPositions["class"].refreshParameters()
        SubPositions["global"] .refreshParameters()

        configApp.updateView()
        writeParameters()

    except:
        ac.log("SubStanding: Error in refreshAndWriteParameters: %s" % traceback.format_exc())

def acUpdate(deltaT):
    try:
        global myDriver, lastUpdateTime, driversFullList

        lastUpdateTime += deltaT

        if lastUpdateTime < float(updateTime)/1000:
            return

        if debug:
            cpuUsage.startMesure()

        lastUpdateTime = 0

        modelStanding = []
        modelCount = 0
        myModelPos = 0

        classStanding = []
        classCount = 0
        myClassPos = 0

        globalStanding = []
        globalCount = 0
        myGlobalPos = 0

        relativeStanding = []

        # get only connected drivers
        driversList = []

        # Get all needed infos
        for driver in driversFullList:
            index = driver["carId"]
            driver["connected"]   = ac.isConnected(index)

            # jump to next driver if not connected
            if not driver["connected"]:
                continue

            driver["driverName"]  = ac.getDriverName(index)
            driver["carName"]     = ac.getCarName(index)
            driver["bestLap"]     = ac.getCarState(index, acsys.CS.BestLap)
            driver["lapCount"]    = ac.getCarState(index, acsys.CS.LapCount)
            driver["speedMS"]     = ac.getCarState(index, acsys.CS.SpeedMS)
            driver["lapDistance"] = ac.getCarState(index, acsys.CS.NormalizedSplinePosition)
            driver["distance"]    = driver["lapCount"] + driver["lapDistance"]
            driver["tyres"]       = ac.getCarTyreCompound(index)

            isInPitLine = ac.isCarInPitline(index)

            # Race
            if info.graphics.session == 2:
                # Trying to resolve the race start problem
                if driver["raceStartFlag"] and driver["lapCount"] == 0:
                    if driver["distance"] > 0.7:
                        driver["distance"] -= 1.0
                    elif driver["distance"] > 0.0:
                        driver["raceStartFlag"] = False

                    driver["pitStops"] = 0
                else:
                    driver["raceStartFlag"] = False

                # Pit entry
                if isInPitLine and not driver["isInPitLine"]:
                    driver["pitEntryFlag"] = True
                # Pit exit
                elif driver["pitEntryFlag"] and not isInPitLine and driver["isInPitLine"]:
                    driver["pitEntryFlag"] = False
                    driver["pitStops"] += 1
            else:
                driver["pitStops"] = 0
                driver["raceStartFlag"] = True

            driver["isInPitLine"] = isInPitLine

            driversList.append(driver)

            if index == 0:
                myDriver = driver

        driversCount = len(driversList)

        # Sort the global standing by distance or best lap
        # Race
        if info.graphics.session == 2:
            #sort by distance
            driversList.sort(key=lambda tup: tup["distance"], reverse=True)
            globalStanding = driversList
        # Practice, qualif
        else:
            # Only display drivers with a valid best lap
            for eachDriver in driversList:
                if eachDriver["bestLap"] == 0:
                    eachDriver["bestLap"] = 9999999
                globalStanding.append(eachDriver)

            #sort by bestLap
            globalStanding.sort(key=lambda tup: tup["bestLap"])

        allCarsStopped = True
        # Extract sub-standings
        for driverIndex, eachDriver in enumerate(globalStanding):
            eachDriver["position"] = driverIndex

            if allCarsStopped:
                allCarsStopped = eachDriver["speedMS"] <= 0.1

            # sort into model/classStandings
            if eachDriver["carName"] == myCarName:
                modelStanding.append(eachDriver)

            if eachDriver["carName"] in myClass:
                classStanding.append(eachDriver)

        # reset the raceStartFlag assuming a session restart
        if allCarsStopped:
            for eachDriver in globalStanding:
                eachDriver["raceStartFlag"] = True

        # Get relative list
        relativeStanding = driversList.copy()
        relativeStanding.sort(key=lambda source: myDriver["lapDistance"] - source["lapDistance"])
        rotationAmount = relativeStanding.index(myDriver) - (math.ceil(len(relativeStanding) / 2.0) - 1) # current position - target position
        relativeStanding = relativeStanding[rotationAmount:] + relativeStanding[:rotationAmount]

        # Update positions
        modelCount = len(modelStanding)
        if myDriver in modelStanding:
            myModelPos = modelStanding.index(myDriver) + 1
        else:
            myModelPos = 0

        classCount = len(classStanding)
        if myDriver in classStanding:
            myClassPos = classStanding.index(myDriver) + 1
        else:
            myClassPos = 0

        globalCount = len(globalStanding)
        if myDriver in globalStanding:
            myGlobalPos = globalStanding.index(myDriver) + 1
        else:
            myGlobalPos = 0

        # Update the views
        SubStandings["model"].updateView(modelStanding)
        SubStandings["class"].updateView(classStanding)
        SubStandings["global"].updateView(globalStanding)
        SubStandings["relative"].updateView(relativeStanding)

        # update all bottom anchor positions and rewrite config if necessary
        updateBottomAnchorPositions()

        SubPositions["model"].updateView(myModelPos, modelCount)
        SubPositions["class"].updateView(myClassPos, classCount)
        SubPositions["global"].updateView(myGlobalPos, globalCount)

        configApp.updateView()

        if debug:
            cpuUsage.stopMesure()
            cpuUsage.updateView()

    except:
        ac.log("SubStanding: Error in acUpdate: %s" % traceback.format_exc())

def onRenderCallbackStandingModel(deltaT):
    try:
        SubStandings["model"].onRenderCallback(deltaT)
    except:
        ac.log("SubStanding: Error in onRenderCallbackStandingModel: %s" % traceback.format_exc())

def onRenderCallbackStandingClass(deltaT):
    try:
        SubStandings["class"].onRenderCallback(deltaT)
    except:
        ac.log("SubStanding: Error in onRenderCallbackStandingClass: %s" % traceback.format_exc())

def onRenderCallbackStandingGlobal(deltaT):
    try:
        SubStandings["global"].onRenderCallback(deltaT)
    except:
        ac.log("SubStanding: Error in onRenderCallbackStandingGlobal: %s" % traceback.format_exc())

def onRenderCallbackStandingRelative(deltaT):
    try:
        SubStandings["relative"].onRenderCallback(deltaT)
    except:
        ac.log("SubStanding: Error in onRenderCallbackStandingRelative: %s" % traceback.format_exc())

def onRenderCallbackPositionModel(deltaT):
    try:
        SubPositions["model"].onRenderCallback(deltaT)
    except:
        ac.log("SubStanding: Error in onRenderCallbackPositionModel: %s" % traceback.format_exc())

def onRenderCallbackPositionClass(deltaT):
    try:
        SubPositions["class"].onRenderCallback(deltaT)
    except:
        ac.log("SubStanding: Error in onRenderCallbackPositionClass: %s" % traceback.format_exc())

def onRenderCallbackPositionGlobal(deltaT):
    try:
        SubPositions["global"].onRenderCallback(deltaT)
    except:
        ac.log("SubStanding: Error in onRenderCallbackPositionGlobal: %s" % traceback.format_exc())

def onRenderCallbackConfig(deltaT):
    try:
        configApp.onRenderCallback(deltaT)
    except:
        ac.log("SubStanding: Error in onRenderCallbackConfig: %s" % traceback.format_exc())

class StandingWindow:

    def __init__(self, name, headerName, driverDisplayed, isRelative):
        try:
            self.name = name
            self.headerName = headerName
            self.window = ac.newApp(name)
            self.isActive = False

            ac.addOnAppActivatedListener(self.window, standingActivated)
            ac.addOnAppDismissedListener(self.window, standingDismissed)

            self.positionLabel = []
            self.nameLabel = []
            self.tyresLabel = []
            self.pitLabel = []
            self.timeLabel = []
            self.carIdDisplayed = []
            self.currentDriverDisplayed = 1
            self.isRelative = isRelative

            self.width = -1
            self.height = -1

            for index in range(maxDriverCount):
                self.positionLabel.append(ac.addLabel(self.window, ""))
                ac.setFontAlignment(self.positionLabel[index], 'left')

                self.nameLabel.append(ac.addLabel(self.window, ""))
                ac.setFontAlignment(self.nameLabel[index], 'left')

                self.tyresLabel.append(ac.addLabel(self.window, ""))
                ac.setFontAlignment(self.tyresLabel[index], 'center')

                self.pitLabel.append(ac.addLabel(self.window, ""))
                ac.setFontAlignment(self.pitLabel[index], 'center')

                self.timeLabel.append(ac.addLabel(self.window, ""))
                ac.setFontAlignment(self.timeLabel[index], 'right')

                self.carIdDisplayed.append(-1)

            self.refreshParameters(driverDisplayed)
        except:
            ac.log("SubStanding: Error in StandingWindow constructor: %s" % traceback.format_exc())

    def refreshParameters(self, driverDisplayed):
        self.driverDisplayed = driverDisplayed

        if showLogo:
            ac.setIconPosition(self.window, 0, 0)
        else:
            ac.setIconPosition(self.window, -10000, -10000)

        if showTitle:
            ac.setTitle(self.window, self.headerName)
        else:
            ac.setTitle(self.window, "")

        if showLogo or showTitle:
            self.firstSpacing = firstSpacing
        else:
            self.firstSpacing = 0

        self.widthPos   = fontSize*2
        self.widthBadge = showBadge*(fontSize + spacing)
        self.widthName  = fontSize*10
        self.widthTyres = showTyres*fontSize*2
        self.widthPitStops = showPitStops*fontSize*2
        self.widthTime  = showDelta*fontSize*5 + self.isRelative*showLapDiffRelative*fontSize*3

        width      = self.widthPos + self.widthBadge + self.widthName + self.widthTyres + self.widthPitStops + self.widthTime + 2*spacing
        height     = self.firstSpacing + (fontSize + spacing)*self.currentDriverDisplayed + spacing

        for index in range(maxDriverCount):
            ac.setFontSize(self.positionLabel[index], fontSize)
            ac.setPosition(self.positionLabel[index], spacing, self.firstSpacing + index*(fontSize+spacing))
            ac.setSize(self.positionLabel[index], self.widthPos/2+3, fontSize+spacing)

            ac.setFontSize(self.nameLabel[index], fontSize)
            ac.setPosition(self.nameLabel[index], spacing + self.widthPos + self.widthBadge, self.firstSpacing + index*(fontSize+spacing))
            ac.setSize(self.nameLabel[index], self.widthName, fontSize+spacing)

            ac.setFontSize(self.tyresLabel[index], fontSize)
            ac.setPosition(self.tyresLabel[index], spacing + self.widthPos + self.widthBadge + self.widthName, self.firstSpacing + index*(fontSize+spacing))
            ac.setSize(self.tyresLabel[index], self.widthTyres, fontSize+spacing)
            ac.setVisible(self.tyresLabel[index], showTyres)

            ac.setFontSize(self.pitLabel[index], fontSize)
            ac.setPosition(self.pitLabel[index], spacing + self.widthPos + self.widthBadge + self.widthName + self.widthTyres, self.firstSpacing + index*(fontSize+spacing))
            ac.setSize(self.pitLabel[index], self.widthPitStops, fontSize+spacing)
            ac.setVisible(self.pitLabel[index], showPitStops)

            ac.setFontSize(self.timeLabel[index], fontSize)
            ac.setPosition(self.timeLabel[index], spacing + self.widthPos + self.widthBadge + self.widthName + self.widthTyres + self.widthPitStops, self.firstSpacing + index*(fontSize+spacing))
            ac.setSize(self.timeLabel[index], self.widthTime, fontSize+spacing)
            ac.setVisible(self.timeLabel[index], showDelta)

        # Adjust window size, opacity and border
        self.setSize(width, height)

    def setSize(self, width, height):
        if height != self.height or width != self.width:
            self.width = width
            self.height = height
            ac.setSize(self.window, width, height)
            # reposition so that bottom is where it was before
            if bottomAnchor:
                ac.setPosition(self.window, ac.getPosition(self.window)[0], BottomAnchors[self.name] - height)
            ac.setBackgroundOpacity(self.window, float(opacity)/100)
            ac.drawBorder(self.window, showBorder)

    def updateBottomAnchorPosition(self):
        bottomAnchorPosition = int(ac.getPosition(self.window)[1] + self.height)
        if bottomAnchorPosition != BottomAnchors[self.name]:
            BottomAnchors[self.name] = int(ac.getPosition(self.window)[1] + self.height)
            return True
        return False

    def updateView(self, standing):
        # cancel update if inactive
        if not self.isActive:
            return

        try:
            myIndex = standing.index(myDriver)
            # Less drivers or equal than the max, display all of them
            if len(standing) <= self.driverDisplayed:
                driverIndex = 0
                lastIndex = len(standing) - 1

                self.currentDriverDisplayed = len(standing)
            # Only display "driverDisplayed" drivers
            else:
                self.currentDriverDisplayed = self.driverDisplayed

                # We are on the top places of the standing, display from first
                if myIndex <= (self.driverDisplayed - 1)/2:
                    driverIndex = 0
                    lastIndex = self.driverDisplayed - 1
                # We are on the bottom places of the standing, display until last
                elif myIndex >= len(standing) - (self.driverDisplayed - 1)/2:
                    driverIndex = len(standing) - self.driverDisplayed
                    lastIndex = len(standing) - 1
                # We are on the middle of the standing
                else:
                    firstIndex = int(myIndex - (self.driverDisplayed - 1)/2)

                    if firstIndex < 0:
                        firstIndex = 0

                    lastIndex = firstIndex + self.driverDisplayed - 1

                    if lastIndex >= len(standing):
                        lastIndex = len(standing)-1

                    driverIndex = firstIndex


            for labelIndex in range(maxDriverCount):
                if driverIndex <= lastIndex:
                    driver = standing[driverIndex]

                    ac.setVisible(self.positionLabel[labelIndex], 1)
                    ac.setVisible(self.nameLabel[labelIndex], 1)
                    ac.setVisible(self.tyresLabel[labelIndex], showTyres)
                    ac.setVisible(self.pitLabel[labelIndex], showPitStops)
                    ac.setVisible(self.timeLabel[labelIndex], showDelta)

                    # Refresh position and name
                    ac.setText(self.positionLabel[labelIndex], "%i." % (driver["position"]+1))
                    ac.setText(self.nameLabel[labelIndex], driver["driverName"])
                    self.carIdDisplayed[labelIndex] = driver["carId"]

                    # Refresh time/distance
                    # if Racing (or in relative mode)
                    if info.graphics.session == 2 or self.isRelative:
                        # My car
                        if driver["carId"] == 0:
                            if myDriver["lapCount"] > 1:
                                ac.setText(self.timeLabel[labelIndex], "%i laps" % myDriver["lapCount"])
                            else:
                                ac.setText(self.timeLabel[labelIndex], "%i lap" % myDriver["lapCount"])
                        # Other car
                        else:
                            # Delta string and color
                            self.updateRow(labelIndex, driver, driverIndex - myIndex)
                    # Practice, qualif
                    else:
                        ac.setText(self.timeLabel[labelIndex], timeToString(driver["bestLap"]))
                        ac.setFontColor(self.timeLabel[labelIndex], 1, 1, 1, 1)

                    # Update row color for my car
                    if driver["carId"] == 0:
                        # Color my car in light blue
                        self.setRowColor(labelIndex, 0, 1, 1, 1)
                        ac.setFontColor(self.timeLabel[labelIndex], 0, 1, 1, 1)

                    ac.setText(self.tyresLabel[labelIndex], driver["tyres"])

                    if driver["isInPitLine"]:
                        ac.setText(self.pitLabel[labelIndex], "PIT")
                    elif info.graphics.session == 2:
                        ac.setText(self.pitLabel[labelIndex], "%ip" % driver["pitStops"])
                    else:
                        ac.setText(self.pitLabel[labelIndex], "")

                    driverIndex += 1
                else:
                    ac.setVisible(self.positionLabel[labelIndex], 0)
                    ac.setVisible(self.nameLabel[labelIndex], 0)
                    ac.setVisible(self.tyresLabel[labelIndex], 0)
                    ac.setVisible(self.pitLabel[labelIndex], 0)
                    ac.setVisible(self.timeLabel[labelIndex], 0)
                    ac.setText(self.timeLabel[labelIndex], "")
                    self.carIdDisplayed[labelIndex] = -1

            # Adjust window size, opacity and border
            self.setSize(self.width, self.firstSpacing + (fontSize + spacing)*self.currentDriverDisplayed + spacing)

        except:
            ac.log("SubStanding: Error in StandingWindow.updateView: %s" % traceback.format_exc())

    def updateRow(self, labelIndex, driver, indexDiff):
        if self.isRelative > 0:
            self.updateRowDistanceRelative(labelIndex, driver, indexDiff)
        else:
            self.updateRowDistance(labelIndex, driver)

    # global distance calculation
    def updateRowDistance(self, labelIndex, driver):
        # calculate values
        delta = driver["distance"] - myDriver["distance"]

        if delta > 100:
            delta = 0

        speedAvg = (driver["speedMS"] + myDriver["speedMS"])/2
        distance = delta*trackLength

        # set relative distance
        distanceInLap = distance
        if trackLength > 0:
            distanceInLap = abs(distance) % trackLength

        timeLabel = self.timeLabel[labelIndex]

        # update label text
        if delta < 1 and delta > -1 and trackLength > 0:
            if unit == "metric":
                ac.setText(timeLabel, "%+.0f m" % distance)
            elif unit == "imperial":
                if distance < 1609:
                    ac.setText(timeLabel, "%+.0f ft" % (distance*3.28))
                else:
                    ac.setText(timeLabel, "%+.1f mi" % (distance/1609))
            elif unit == "time":
                if speedAvg < 15:
                    speedAvg = 15
                deltaTime = distance/speedAvg
                ac.setText(timeLabel, "%+.1f s" % deltaTime)

        elif delta < 2 and delta > -2:
            ac.setText(timeLabel, "%+.1f lap" % delta)
        else:
            ac.setText(timeLabel, "%+.1f laps" % delta)

        # update color
        self.updateRowDistanceColor(timeLabel, math.copysign(distanceInLap, distance), int(delta))
        self.updateRowColor(labelIndex, driver, int(delta))

    # relative distance calculation
    def updateRowDistanceRelative(self, labelIndex, driver, indexDiff):
        isRace = info.graphics.session == 2
        delta = 0.0
        if indexDiff < 0: # other driver above player in list
            delta = driver["lapDistance"] - myDriver["lapDistance"]
        else: # other player below in player list
            delta = myDriver["lapDistance"] - driver["lapDistance"]

        # lap difference calc
        lapDiff = 0
        if isRace: # calculate lap difference only in race mode
            driverLapCount = driver["lapCount"] - (1 if driver["raceStartFlag"] and driver["lapCount"] == 0 else 0)
            myDriverLapCount = myDriver["lapCount"] - (1 if myDriver["raceStartFlag"] and myDriver["lapCount"] == 0 else 0)
            lapDiff = math.copysign(myDriverLapCount - driverLapCount, myDriver["position"] - driver["position"])

        if delta < 0:
            delta += 1
            if isRace:
                lapDiff += math.copysign(1, indexDiff)

        speedAvg = (driver["speedMS"] + myDriver["speedMS"])/2
        distance = math.copysign(delta*trackLength, -indexDiff)

        timeLabel = self.timeLabel[labelIndex]

        # update label text
        labelText = "%+.0f m" % distance # default metric
        if unit == "imperial":
            if distance < 1609:
                labelText = "%+.0f ft" % (distance*3.28)
            else:
                labelText = "%+.1f mi" % (distance/1609)
        elif unit == "time":
            if speedAvg < 15:
                speedAvg = 15
            deltaTime = distance/speedAvg
            labelText = "%+.1f s" % deltaTime
        ac.setText(timeLabel, labelText + ((" (%+.0f)" % lapDiff) if showLapDiffRelative and abs(lapDiff) > 0 else ""))

        # update color
        self.updateRowDistanceColor(timeLabel, distance, lapDiff)
        self.updateRowColor(labelIndex, driver, lapDiff)

    def updateRowDistanceColor(self, timeLabel, distance, lapDiff):
        # update color
        # colorFactor
        colorFactor = 1.0
        if(colorAt > 0):
            colorFactor = 1-max(min(float(abs(distance))/colorAt, 1.0), 0.0) # 0 if > colorAt, 1 if distance == 0, clamped to [0,1]

        if colorFactor <= 0.0:
            # White (only fallback)
            ac.setFontColor(timeLabel, 1, 1, 1, 1)
        if lapDiff > 0:
            # Yellow (car is > 1 lap in front)
            ac.setFontColor(timeLabel, 1, 1, 1-colorFactor, 1)
        elif lapDiff < 0:
            # Blue (car is > 1 lap in behind)
            ac.setFontColor(timeLabel, 1-colorFactor, 1-(0.5*colorFactor), 1, 1)
        elif distance > 0:
            # Orange
            ac.setFontColor(timeLabel, 1, 1-(0.5*colorFactor), 1-colorFactor, 1)
        elif distance < 0:
            # Turkish
            ac.setFontColor(timeLabel, 1-colorFactor, 1, 1-(0.5*colorFactor), 1)
        else:
            # White (only fallback)
            ac.setFontColor(timeLabel, 1, 1, 1, 1)

    def updateRowColor(self, labelIndex, driver, lapDiff):
        alpha = 1 if lapDiff == 0 else 0.6

        # Position and name in blue for the teammates, pink for the rivals and white for the others
        if driver["driverName"] in teammates:
            self.setRowColor(labelIndex, 0, 1, 1, alpha)
        elif driver["driverName"] in rivals:
            self.setRowColor(labelIndex, 1, 0, 1, alpha)
        else:
            self.setRowColor(labelIndex, 1, 1, 1, alpha)

    def setRowColor(self, labelIndex, colorR, colorG, colorB, colorAlpha):
        ac.setFontColor(self.positionLabel[labelIndex], colorR, colorG, colorB, colorAlpha)
        ac.setFontColor(self.nameLabel[labelIndex], colorR, colorG, colorB, colorAlpha)
        ac.setFontColor(self.tyresLabel[labelIndex], colorR, colorG, colorB, colorAlpha)
        ac.setFontColor(self.pitLabel[labelIndex], colorR, colorG, colorB, colorAlpha)

    def onActivated(self):
        self.isActive = True
        ac.log("SubStanding: Activated %s" % self.name)

    def onDismissed(self):
        self.isActive = False
        ac.log("SubStanding: Dismissed %s" % self.name)

    def onRenderCallback(self, deltaT):
        if not self.isActive:
            self.onActivated()

        for i in range(maxDriverCount):
            if self.carIdDisplayed[i] >= 0 and badge[self.carIdDisplayed[i]] > 0 and showBadge:
                ac.glColor4f(1,1,1,1)
                ac.glQuadTextured(self.widthPos, self.firstSpacing + i*(fontSize+spacing) + spacing, fontSize, fontSize, badge[self.carIdDisplayed[i]])
            else:
                ac.glColor4f(1,1,1,0)

        ac.setBackgroundOpacity(self.window, float(opacity)/100)
        ac.drawBorder(self.window, showBorder)

# SubStanding related functions

def updateBottomAnchorPositions():
    if (bottomAnchor and
        (SubStandings["global"].updateBottomAnchorPosition() or
            SubStandings["relative"].updateBottomAnchorPosition() or
            SubStandings["model"].updateBottomAnchorPosition() or
            SubStandings["class"].updateBottomAnchorPosition())):
        refreshAndWriteParameters()

def standingActivated(app):
    for name in SubStandings:
        subStanding = SubStandings[name]
        if subStanding.window == app:
            subStanding.onActivated()

def standingDismissed(app):
    for name in SubStandings:
        subStanding = SubStandings[name]
        if subStanding.window == app:
            subStanding.onDismissed()


class PositionWindow:
    def __init__(self, name, headerName):
        self.headerName = headerName
        self.window = ac.newApp(name)
        self.isActive = False

        ac.addOnAppActivatedListener(self.window, positionActivated)
        ac.addOnAppDismissedListener(self.window, positionDismissed)

        ac.setTitle(self.window, headerName)

        self.label = ac.addLabel(self.window, "x/x")
        ac.setFontAlignment(self.label, 'center')

        self.refreshParameters()

    def refreshParameters(self):
        if showLogoPos:
            ac.setIconPosition(self.window, 0, 0)
        else:
            ac.setIconPosition(self.window, -10000, -10000)

        if showTitlePos:
            ac.setTitle(self.window, self.headerName)
        else:
            ac.setTitle(self.window, "")

        width  = fontSizePos*2 + 80
        height = fontSizePos + 50

        ac.setSize(self.window, width, height)

        ac.setPosition(self.label, 0, height/2 - fontSizePos/2)
        ac.setSize(self.label, width, fontSizePos)
        ac.setFontSize(self.label, fontSizePos)

    def updateView(self, myPos, carCount):
        if not self.isActive:
            return

        ac.setText(self.label, "%d/%d" % (myPos, carCount))

    def onActivated(self):
        self.isActive = True
        ac.log("SubStanding: Activated %s" % self.name)

    def onDismissed(self):
        self.isActive = False
        ac.log("SubStanding: Dismissed %s" % self.name)

    def onRenderCallback(self, deltaT):
        if not self.isActive:
            self.onActivated()

        ac.setBackgroundOpacity(self.window, float(opacityPos)/100)
        ac.drawBorder(self.window, showBorderPos)

def positionActivated(app):
    for name in SubPositions:
        subPosition = SubPositions[name]
        if subPosition.window == app:
            subPosition.onActivated()

def positionDismissed(app):
    for name in SubPositions:
        subPosition = SubPositions[name]
        if subPosition.window == app:
            subPosition.onDismissed()

def timeToString(time):
    if time <= 0 or time == 9999999:
        return "-:--.---"
    else:
        return "%d:%02d.%03d" % (int(time/60000), int((time%60000)/1000), time%1000)



'''
Standings:
Show AC logo:       Yes     Change
Show title:         Yes     Change
Font size:          18      + -
Opacity:            50 %    + -
Anchor bottom:      No      Change
Show border:        Yes     Change
Show badges:        Yes     Change
Show tyres:         Yes     Change
Show pit stops:     Yes     Change
Show deltas:        Yes     Change
Show laps in Rel.:  No      Change
Units:              Metric  Change
Color distance:     250 m   + -
Max drivers:
Global:             All     + -
Relative:           All     + -
Class:              All     + -
Same car:           All     + -

Positions:
Show AC logo:       Yes     Change
Show title:         Yes     Change
Font size:          18      + -
Opacity:            50 %    + -
Show border:        Yes     Change

Both:
Refresh every:      0.2 s   + -
'''
class SubStanding_config:

    def __init__(self, name, headerName, fontSize):
        self.window = ac.newApp(name)
        self.isActive = False

        ac.addOnAppActivatedListener(self.window, configActivated)
        ac.addOnAppDismissedListener(self.window, configDismissed)

        ac.setTitle(self.window, headerName)

        numberOfItems   = 29
        widthLeft       = fontSize*10
        widthCenter     = fontSize*5
        widthRight      = fontSize*5
        width           = widthLeft + widthCenter + widthRight + 2*spacing
        height          = firstSpacing + (fontSize*1.5 + spacing)*numberOfItems

        ac.setSize(self.window, width, height)

        self.leftLabel = []
        self.centerLabel = []
        self.changeButton = []
        self.plusButton = []
        self.minusButton = []

        for index in range(numberOfItems):
            self.leftLabel.append(ac.addLabel(self.window, ""))
            ac.setFontSize(self.leftLabel[index], fontSize)
            ac.setPosition(self.leftLabel[index], spacing, firstSpacing + index*(fontSize*1.5+spacing))
            ac.setSize(self.leftLabel[index], widthLeft, fontSize+spacing)
            ac.setFontAlignment(self.leftLabel[index], 'left')

            self.centerLabel.append(ac.addLabel(self.window, ""))
            ac.setFontSize(self.centerLabel[index], fontSize)
            ac.setPosition(self.centerLabel[index], spacing + widthLeft, firstSpacing + index*(fontSize*1.5+spacing))
            ac.setSize(self.centerLabel[index], widthCenter, fontSize+spacing)
            ac.setFontAlignment(self.centerLabel[index], 'left')

            self.changeButton.append(ac.addButton(self.window, "Change"))
            ac.setFontSize(self.changeButton[index], fontSize)
            ac.setPosition(self.changeButton[index], spacing + widthLeft + widthCenter, firstSpacing + index*(fontSize*1.5+spacing))
            ac.setSize(self.changeButton[index], fontSize*4, fontSize*1.5)

            self.plusButton.append(ac.addButton(self.window, "+"))
            ac.setFontSize(self.plusButton[index], fontSize)
            ac.setPosition(self.plusButton[index], spacing + widthLeft + widthCenter, firstSpacing + index*(fontSize*1.5+spacing))
            ac.setSize(self.plusButton[index], fontSize*1.5, fontSize*1.5)

            self.minusButton.append(ac.addButton(self.window, "-"))
            ac.setFontSize(self.minusButton[index], fontSize)
            ac.setPosition(self.minusButton[index], spacing + widthLeft + widthCenter + fontSize*2.5, firstSpacing + index*(fontSize*1.5+spacing))
            ac.setSize(self.minusButton[index], fontSize*1.5, fontSize*1.5)

        rowIndex = 0

        ac.setText(self.leftLabel[rowIndex], "Standings:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Show AC logo:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleLogo)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showLogoId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Show title:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleTitle)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showTitleId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Font size:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.addOnClickedListener(self.plusButton[rowIndex], fontSizePlus)
        ac.addOnClickedListener(self.minusButton[rowIndex], fontSizeMinus)
        self.fontSizeId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Opacity:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.addOnClickedListener(self.plusButton[rowIndex], opacityPlus)
        ac.addOnClickedListener(self.minusButton[rowIndex], opacityMinus)
        self.opacityId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Anchor bottom:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleBottomAnchor)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.bottomAnchorId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Show border:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleBorder)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showBorderId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Show badges:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleBadge)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showBadgeId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Show tyres:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleTyres)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showTyresId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Show pit stops:")
        ac.addOnClickedListener(self.changeButton[rowIndex], togglePitStops)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showPitStopsId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Show delta:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleDelta)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showDeltaId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Show Laps in Rel.:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleLapDiffRelative)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showLapDiffRelativeId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Units:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleUnits)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.unitsId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Color distance:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.addOnClickedListener(self.plusButton[rowIndex], colorAtPlus)
        ac.addOnClickedListener(self.minusButton[rowIndex], colorAtMinus)
        self.colorAtId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Max drivers:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "- Global:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.addOnClickedListener(self.plusButton[rowIndex], driverGlobalPlus)
        ac.addOnClickedListener(self.minusButton[rowIndex], driverGlobalMinus)
        self.driverGlobalId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "- Relative:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.addOnClickedListener(self.plusButton[rowIndex], driverRelativePlus)
        ac.addOnClickedListener(self.minusButton[rowIndex], driverRelativeMinus)
        self.driverRelativeId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "- Class:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.addOnClickedListener(self.plusButton[rowIndex], driverClassPlus)
        ac.addOnClickedListener(self.minusButton[rowIndex], driverClassMinus)
        self.driverClassId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "- Same car:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.addOnClickedListener(self.plusButton[rowIndex], driverSameCarPlus)
        ac.addOnClickedListener(self.minusButton[rowIndex], driverSameCarMinus)
        self.driverSameCarId = rowIndex

        rowIndex += 1

        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Positions:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Show AC logo:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleLogoPos)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showLogoPosId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Show title:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleTitlePos)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showTitlePosId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Font size:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.addOnClickedListener(self.plusButton[rowIndex], fontSizePosPlus)
        ac.addOnClickedListener(self.minusButton[rowIndex], fontSizePosMinus)
        self.fontSizePosId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Opacity:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.addOnClickedListener(self.plusButton[rowIndex], opacityPosPlus)
        ac.addOnClickedListener(self.minusButton[rowIndex], opacityPosMinus)
        self.opacityPosId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Show border:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleBorderPos)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showBorderPosId = rowIndex

        rowIndex += 1

        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Global:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Refresh every:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.addOnClickedListener(self.plusButton[rowIndex], refreshPlus)
        ac.addOnClickedListener(self.minusButton[rowIndex], refreshMinus)
        self.refreshId = rowIndex

    def updateView(self):
        if not self.isActive:
            return

        ac.setText(self.centerLabel[self.showLogoId],  yesOrNo(showLogo))
        ac.setText(self.centerLabel[self.showTitleId], yesOrNo(showTitle))
        ac.setText(self.centerLabel[self.fontSizeId],  str(fontSize))
        ac.setText(self.centerLabel[self.opacityId],   "%i %%" % opacity)
        ac.setText(self.centerLabel[self.bottomAnchorId], yesOrNo(bottomAnchor))
        ac.setText(self.centerLabel[self.showBorderId], yesOrNo(showBorder))
        ac.setText(self.centerLabel[self.showTyresId], yesOrNo(showTyres))
        ac.setText(self.centerLabel[self.showPitStopsId], yesOrNo(showPitStops))
        ac.setText(self.centerLabel[self.showBadgeId], yesOrNo(showBadge))
        ac.setText(self.centerLabel[self.showDeltaId], yesOrNo(showDelta))
        ac.setText(self.centerLabel[self.showLapDiffRelativeId], yesOrNo(showLapDiffRelative))

        if unit == "metric":
            ac.setText(self.centerLabel[self.unitsId], "Metric")
            ac.setText(self.centerLabel[self.colorAtId], "%i m" % colorAt)
        elif unit == "imperial":
            ac.setText(self.centerLabel[self.unitsId], "Imperial")
            ac.setText(self.centerLabel[self.colorAtId], "%.0f ft" % (colorAt*3.28))
        elif unit == "time":
            ac.setText(self.centerLabel[self.unitsId], "Time (est)")
            ac.setText(self.centerLabel[self.colorAtId], "%i m" % colorAt)

        ac.setText(self.centerLabel[self.driverGlobalId], str(driverDisplayedGlobal))
        ac.setText(self.centerLabel[self.driverRelativeId], str(driverDisplayedRelative))
        ac.setText(self.centerLabel[self.driverClassId], str(driverDisplayedClass))
        ac.setText(self.centerLabel[self.driverSameCarId], str(driverDisplayedSameCar))

        ac.setText(self.centerLabel[self.showLogoPosId],  yesOrNo(showLogoPos))
        ac.setText(self.centerLabel[self.showTitlePosId], yesOrNo(showTitlePos))
        ac.setText(self.centerLabel[self.fontSizePosId],  str(fontSizePos))
        ac.setText(self.centerLabel[self.opacityPosId],   "%i %%" % opacityPos)
        ac.setText(self.centerLabel[self.showBorderPosId], yesOrNo(showBorderPos))

        if updateTime == 0:
            ac.setText(self.centerLabel[self.refreshId], "Min")
        elif updateTime == 50:
            ac.setText(self.centerLabel[self.refreshId], "0.05 s")
        else:
            ac.setText(self.centerLabel[self.refreshId], "%.1f s" % (float(updateTime)/1000))

    def onRenderCallback(self, deltaT):
        if not self.isActive:
            self.onActivated()
        ac.setBackgroundOpacity(self.window, 1.0)

    def onActivated(self):
        self.isActive = True
        ac.log("SubStanding: Activated Config")

    def onDismissed(self):
        self.isActive = False
        ac.log("SubStanding: Dismissed Config")

    def onRenderCallback(self, deltaT):

        ac.setBackgroundOpacity(self.window, float(opacityPos)/100)
        ac.drawBorder(self.window, showBorderPos)

def configActivated(app):
    configApp.onActivated()

def configDismissed(app):
    configApp.onDismissed()

def yesOrNo(value):
    if value:
        return "Yes"
    else:
        return "No"

def toggleLogo(dummy, variable):
    global showLogo

    if showLogo:
        showLogo = 0
    else:
        showLogo = 1

    refreshAndWriteParameters()

def toggleTitle(dummy, variable):
    global showTitle

    if showTitle:
        showTitle = 0
    else:
        showTitle = 1

    refreshAndWriteParameters()

def fontSizePlus(dummy, variable):
    global fontSize

    fontSize += 1

    refreshAndWriteParameters()

def fontSizeMinus(dummy, variable):
    global fontSize

    if fontSize > 6:
        fontSize -= 1

    refreshAndWriteParameters()

def toggleBottomAnchor(dummy, variable):
    global bottomAnchor

    if bottomAnchor:
        bottomAnchor = 0
    else:
        bottomAnchor = 1
        SubStandings["global"].updateBottomAnchorPosition()
        SubStandings["relative"].updateBottomAnchorPosition()
        SubStandings["model"].updateBottomAnchorPosition()
        SubStandings["class"].updateBottomAnchorPosition()

    refreshAndWriteParameters()

def toggleBorder(dummy, variable):
    global showBorder

    if showBorder:
        showBorder = 0
    else:
        showBorder = 1

    refreshAndWriteParameters()

def toggleTyres(dummy, variable):
    global showTyres

    if showTyres:
        showTyres = 0
    else:
        showTyres = 1

    refreshAndWriteParameters()

def togglePitStops(dummy, variable):
    global showPitStops

    if showPitStops:
        showPitStops = 0
    else:
        showPitStops = 1

    refreshAndWriteParameters()

def toggleBadge(dummy, variable):
    global showBadge

    if showBadge:
        showBadge = 0
    else:
        showBadge = 1

    refreshAndWriteParameters()

def opacityPlus(dummy, variable):
    global opacity

    if opacity < 100:
        opacity += 10

    refreshAndWriteParameters()

def opacityMinus(dummy, variable):
    global opacity

    if opacity >= 10:
        opacity -= 10

    refreshAndWriteParameters()

def toggleDelta(dummy, variable):
    global showDelta

    if showDelta:
        showDelta = 0
    else:
        showDelta = 1

    refreshAndWriteParameters()

def toggleLapDiffRelative(dummy, variable):
    global showLapDiffRelative

    if showLapDiffRelative:
        showLapDiffRelative = 0
    else:
        showLapDiffRelative = 1

    refreshAndWriteParameters()

def colorAtPlus(dummy, variable):
    global colorAt

    if colorAt < 2000:
        colorAt += 50
        colorAt = min(colorAt, 2000)

    refreshAndWriteParameters()

def colorAtMinus(dummy, variable):
    global colorAt

    if colorAt > 0:
        colorAt -= 50
        colorAt = max(colorAt, 0)

    refreshAndWriteParameters()

def toggleUnits(dummy, variable):
    global unit

    if unit == "metric":
        unit = "imperial"
    elif unit == "imperial":
        unit = "time"
    elif unit == "time":
        unit = "metric"

    refreshAndWriteParameters()

def driverGlobalPlus(dummy, variable):
    global driverDisplayedGlobal

    driverDisplayedGlobal += 2

    refreshAndWriteParameters()

def driverGlobalMinus(dummy, variable):
    global driverDisplayedGlobal

    if driverDisplayedGlobal > 3:
        driverDisplayedGlobal -= 2

    refreshAndWriteParameters()

def driverRelativePlus(dummy, variable):
    global driverDisplayedRelative

    driverDisplayedRelative += 2

    refreshAndWriteParameters()

def driverRelativeMinus(dummy, variable):
    global driverDisplayedRelative

    if driverDisplayedRelative > 3:
        driverDisplayedRelative -= 2

    refreshAndWriteParameters()

def driverClassPlus(dummy, variable):
    global driverDisplayedClass

    driverDisplayedClass += 2

    refreshAndWriteParameters()

def driverClassMinus(dummy, variable):
    global driverDisplayedClass

    if driverDisplayedClass > 3:
        driverDisplayedClass -= 2

    refreshAndWriteParameters()

def driverSameCarPlus(dummy, variable):
    global driverDisplayedSameCar

    driverDisplayedSameCar += 2

    refreshAndWriteParameters()

def driverSameCarMinus(dummy, variable):
    global driverDisplayedSameCar

    if driverDisplayedSameCar > 3:
        driverDisplayedSameCar -= 2

    refreshAndWriteParameters()

def toggleLogoPos(dummy, variable):
    global showLogoPos

    if showLogoPos:
        showLogoPos = 0
    else:
        showLogoPos = 1

    refreshAndWriteParameters()

def toggleTitlePos(dummy, variable):
    global showTitlePos

    if showTitlePos:
        showTitlePos = 0
    else:
        showTitlePos = 1

    refreshAndWriteParameters()

def fontSizePosPlus(dummy, variable):
    global fontSizePos

    fontSizePos += 1

    refreshAndWriteParameters()

def fontSizePosMinus(dummy, variable):
    global fontSizePos

    if fontSizePos > 6:
        fontSizePos -= 1

    refreshAndWriteParameters()

def opacityPosPlus(dummy, variable):
    global opacityPos

    if opacityPos < 100:
        opacityPos += 10

    refreshAndWriteParameters()

def opacityPosMinus(dummy, variable):
    global opacityPos

    if opacityPos >= 10:
        opacityPos -= 10

    refreshAndWriteParameters()

def toggleBorderPos(dummy, variable):
    global showBorderPos

    if showBorderPos:
        showBorderPos = 0
    else:
        showBorderPos = 1

    refreshAndWriteParameters()

def refreshPlus(dummy, variable):
    global updateTime

    if updateTime == 0:
        updateTime = 50
    elif updateTime == 50:
        updateTime = 100
    elif updateTime < 1000:
        updateTime += 100
    else:
        updateTime = 1000

    refreshAndWriteParameters()

def refreshMinus(dummy, variable):
    global updateTime

    if updateTime == 50:
        updateTime = 0
    elif updateTime == 100:
        updateTime = 50
    elif updateTime > 100:
        updateTime -= 100
    else:
        updateTime = 0

    refreshAndWriteParameters()

class CpuUsage:
    def __init__(self, name, headerName, fontSize):
        self.name = name
        self.headerName = headerName
        self.window = ac.newApp(name)
        ac.setTitle(self.window, headerName)

        self.firstSpacing = firstSpacing

        widthLeft       = fontSize*6
        widthRight      = fontSize*5
        self.width      = widthLeft + widthRight + 2*spacing
        height          = self.firstSpacing + (fontSize + spacing)*5

        ac.setSize(self.window, self.width, height)

        self.leftLabel = []
        self.rightLabel = []

        for index in range(5):
            self.leftLabel.append(ac.addLabel(self.window, ""))
            ac.setFontSize(self.leftLabel[index], fontSize)
            ac.setPosition(self.leftLabel[index], spacing, self.firstSpacing + index*(fontSize+spacing))
            ac.setSize(self.leftLabel[index], widthLeft, fontSize+spacing)
            ac.setFontAlignment(self.leftLabel[index], 'left')

            self.rightLabel.append(ac.addLabel(self.window, ""))
            ac.setFontSize(self.rightLabel[index], fontSize)
            ac.setPosition(self.rightLabel[index], spacing + widthLeft, self.firstSpacing + index*(fontSize+spacing))
            ac.setSize(self.rightLabel[index], widthRight, fontSize+spacing)
            ac.setFontAlignment(self.rightLabel[index], 'right')

        ac.setText(self.leftLabel[0], "Avg time:")
        ac.setText(self.leftLabel[1], "Min time:")
        ac.setText(self.leftLabel[2], "Max time:")
        ac.setText(self.leftLabel[3], "Avg period:")
        ac.setText(self.leftLabel[4], "Avg load:")

        self.avg = 0
        self.min = 0
        self.max = 999999
        self.avgperiod = 0
        self.avgload = 0

        self.startTime = 0
        self.lastStartTime = 0
        self.delta = 0

    def startMesure(self):
        self.startTime = time.perf_counter()

        if self.lastStartTime == 0:
            self.lastStartTime = self.startTime

        self.lastPeriod = self.startTime - self.lastStartTime
        self.lastStartTime = self.startTime

    def stopMesure(self):
        self.delta = time.perf_counter() - self.startTime

        if self.avg == 0:
            self.avg = self.delta
            self.min = self.delta
            self.max = self.delta
        else:
            self.avg = self.avg*0.9 + self.delta*0.1
            if self.delta < self.min:
                self.min = self.delta
            elif self.delta > self.max:
                self.max = self.delta

        if self.avgperiod == 0:
            self.avgperiod = self.lastPeriod
        else:
            self.avgperiod = self.avgperiod*0.9 + self.lastPeriod*0.1

        if self.avgperiod == 0:
            self.avgload = 0
        else:
            self.avgload = self.avg/self.avgperiod*100

    def updateView(self):
        ac.setText(self.rightLabel[0], "%.3f ms" % (self.avg*1000))
        ac.setText(self.rightLabel[1], "%.3f ms" % (self.min*1000))
        ac.setText(self.rightLabel[2], "%.3f ms" % (self.max*1000))
        ac.setText(self.rightLabel[3], "%.3f ms" % (self.avgperiod*1000))
        ac.setText(self.rightLabel[4], "%.3f %%" % self.avgload)
